41#define GREY_EDGE "greyedge"
43#define SQRT3 1.73205080757
47#define MAX_META_DATA 4
54#define INDEX_NORM INDEX_DX
91#define OFFSET(x) offsetof(ColorConstancyContext, x)
92#define FLAGS AV_OPT_FLAG_FILTERING_PARAM|AV_OPT_FLAG_VIDEO_PARAM
94#define GINDX(s, i) ( (i) - ((s) >> 2) )
111 int filtersize =
s->filtersize;
112 int difford =
s->difford;
113 double sigma =
s->sigma;
117 for (
i = 0;
i <= difford; ++
i) {
120 for (;
i >= 0; --
i) {
133 for (
i = 0;
i < filtersize; ++
i) {
134 s->gauss[0][
i] =
exp(- pow(
GINDX(filtersize,
i), 2.) / (2 * sigma * sigma)) / ( sqrt(2 *
M_PI) * sigma );
135 sum1 +=
s->gauss[0][
i];
137 for (
i = 0;
i < filtersize; ++
i) {
138 s->gauss[0][
i] /= sum1;
145 for (
i = 0;
i < filtersize; ++
i) {
146 s->gauss[1][
i] = - (
GINDX(filtersize,
i) / pow(sigma, 2)) *
s->gauss[0][
i];
147 sum1 +=
s->gauss[1][
i] *
GINDX(filtersize,
i);
150 for (
i = 0;
i < filtersize; ++
i) {
151 s->gauss[1][
i] /= sum1;
158 for (
i = 0;
i < filtersize; ++
i) {
159 s->gauss[2][
i] = ( pow(
GINDX(filtersize,
i), 2) / pow(sigma, 4) - 1/pow(sigma, 2) )
161 sum1 +=
s->gauss[2][
i];
165 for (
i = 0;
i < filtersize; ++
i) {
166 s->gauss[2][
i] -= sum1 / (filtersize);
167 sum2 += (0.5 *
GINDX(filtersize,
i) *
GINDX(filtersize,
i) *
s->gauss[2][
i]);
169 for (
i = 0;
i < filtersize ; ++
i) {
170 s->gauss[2][
i] /= sum2;
191 for (
b = 0;
b < nb_buff; ++
b) {
197 for (p = 0; p < nb_planes; ++p) {
215 int nb_buff =
s->difford + 1;
219 for (
b = 0;
b <= nb_buff; ++
b) {
222 sizeof(*td->
data[
b][p]));
223 if (!td->
data[
b][p]) {
232#define CLAMP(x, mx) av_clip((x), 0, (mx-1))
233#define INDX2D(r, c, w) ( (r) * (w) + (c) )
234#define GAUSS(s, sr, sc, sls, sh, sw, g) ( (s)[ INDX2D(CLAMP((sr), (sh)), CLAMP((sc), (sw)), (sls)) ] * (g) )
258 const int filtersize =
s->filtersize;
259 const double *gauss =
s->gauss[ord];
262 for (plane = 0; plane <
NUM_PLANES; ++plane) {
263 const int height =
s->planeheight[plane];
264 const int width =
s->planewidth[plane];
265 const int in_linesize = in->
linesize[plane];
266 double *
dst = td->
data[dst_index][plane];
272 const uint8_t *
src = in->
data[plane];
279 for (
g = 0;
g < filtersize; ++
g) {
287 const double *
src = td->
data[src_index][plane];
294 for (
g = 0;
g < filtersize; ++
g) {
321 const int difford =
s->difford;
324 for (plane = 0; plane <
NUM_PLANES; ++plane) {
325 const int height =
s->planeheight[plane];
326 const int width =
s->planewidth[plane];
328 const int slice_start = (numpixels * jobnr ) / nb_jobs;
329 const int slice_end = (numpixels * (jobnr+1)) / nb_jobs;
337 norm[
i] = sqrt( pow(dx[
i], 2) + pow(dy[
i], 2));
342 norm[
i] = sqrt( pow(dx[
i], 2) + 4 * pow(dxy[
i], 2) + pow(dy[
i], 2) );
364 int src,
int dst,
int dim,
int nb_threads) {
385 int nb_threads =
s->nb_threads;
386 int height =
s->planeheight[1];
387 int width =
s->planewidth[1];
443 int minknorm =
s->minknorm;
444 const uint8_t thresh = 255;
447 for (plane = 0; plane <
NUM_PLANES; ++plane) {
448 const int height =
s->planeheight[plane];
449 const int width =
s->planewidth[plane];
450 const int in_linesize = in->
linesize[plane];
453 const uint8_t *img_data = in->
data[plane];
463 * (img_data[
INDX2D(
r,
c, in_linesize)] < thresh) );
470 * (img_data[
INDX2D(
r,
c, in_linesize)] < thresh) );
491 int minknorm =
s->minknorm;
492 int difford =
s->difford;
493 double *white =
s->white;
494 int nb_jobs =
FFMIN3(
s->planeheight[1],
s->planewidth[1],
s->nb_threads);
509 for (plane = 0; plane <
NUM_PLANES; ++plane) {
511 for (job = 0; job < nb_jobs; ++job) {
516 for (plane = 0; plane <
NUM_PLANES; ++plane) {
518 for (job = 0; job < nb_jobs; ++job) {
521 white[plane] = pow(white[plane], 1./minknorm);
537 double abs_val = pow( pow(light[0], 2.0) + pow(light[1], 2.0) + pow(light[2], 2.0), 0.5);
543 for (plane = 0; plane <
NUM_PLANES; ++plane) {
547 for (plane = 0; plane <
NUM_PLANES; ++plane) {
548 light[plane] = (light[plane] / abs_val);
574 s->white[0],
s->white[1],
s->white[2]);
577 s->white[0],
s->white[1],
s->white[2]);
600 for (plane = 0; plane <
NUM_PLANES; ++plane) {
601 const int height =
s->planeheight[plane];
602 const int width =
s->planewidth[plane];
604 const int slice_start = (numpixels * jobnr) / nb_jobs;
605 const int slice_end = (numpixels * (jobnr+1)) / nb_jobs;
606 const uint8_t *
src = in->
data[plane];
607 uint8_t *
dst =
out->data[plane];
631 int nb_jobs =
FFMIN3(
s->planeheight[1],
s->planewidth[1],
s->nb_threads);
643 const double break_off_sigma = 3.0;
644 double sigma =
s->sigma;
647 if (!
floor(break_off_sigma * sigma + 0.5) &&
s->difford) {
652 s->filtersize = 2 *
floor(break_off_sigma * sigma + 0.5) + 1;
659 s->planewidth[0] =
s->planewidth[3] = inlink->
w;
661 s->planeheight[0] =
s->planeheight[3] = inlink->
h;
702 int difford =
s->difford;
705 for (
i = 0;
i <= difford; ++
i) {
731 .p.priv_class = &greyedge_class,
uint8_t ptrdiff_t const uint8_t ptrdiff_t int intptr_t intptr_t int int16_t * dst
const FFFilter ff_vf_greyedge
static AVFormatContext * ctx
int ff_filter_frame(AVFilterLink *link, AVFrame *frame)
Send a frame of data to the next filter.
int ff_filter_execute(AVFilterContext *ctx, avfilter_action_func *func, void *arg, int *ret, int nb_jobs)
int ff_filter_get_nb_threads(AVFilterContext *ctx)
Get number of threads for current filter instance.
Main libavfilter public API header.
#define i(width, name, range_min, range_max)
#define AV_CEIL_RSHIFT(a, b)
static __device__ float fabs(float a)
static __device__ float floor(float a)
static int filter_frame(DBEDecodeContext *s, AVFrame *frame)
@ AV_OPT_TYPE_INT
Underlying C type is int.
@ AV_OPT_TYPE_DOUBLE
Underlying C type is double.
#define AVFILTER_FLAG_SUPPORT_TIMELINE_GENERIC
Some filters support a generic "enable" expression option that can be used to enable or disable a fil...
#define AVFILTER_FLAG_SLICE_THREADS
The filter supports multithreading by splitting frames into multiple parts and processing them concur...
int av_frame_is_writable(AVFrame *frame)
Check if the frame data is writable.
void av_frame_free(AVFrame **frame)
Free the frame and any dynamically allocated objects in it, e.g.
int av_frame_copy_props(AVFrame *dst, const AVFrame *src)
Copy only "metadata" fields from src to dst.
#define AV_LOG_TRACE
Extremely verbose debugging, useful for libav* development.
#define AV_LOG_DEBUG
Stuff which is only useful for libav* developers.
#define AV_LOG_ERROR
Something went wrong and cannot losslessly be recovered.
static av_cold void uninit(AVBitStreamFilterContext *ctx)
#define FILTER_INPUTS(array)
#define FILTER_OUTPUTS(array)
static int ff_slice_pos(int total, int jobnr, int nb_jobs)
Compute the boundary index for a slice when work of size total is split into nb_jobs slices.
#define FILTER_SINGLE_PIXFMT(pix_fmt_)
#define AVFILTER_DEFINE_CLASS(fname)
#define NULL_IF_CONFIG_SMALL(x)
Return NULL if CONFIG_SMALL is true, otherwise the argument without modification.
void * av_calloc(size_t nmemb, size_t size)
Memory handling functions.
static int slice_end(AVCodecContext *avctx, AVFrame *pict, int *got_output)
Handle slice ends.
const AVPixFmtDescriptor * av_pix_fmt_desc_get(enum AVPixelFormat pix_fmt)
@ AV_PIX_FMT_GBRP
planar GBR 4:4:4 24bpp
static int config_props(AVBitStreamFilterLink *link)
Describe the class of an AVClass context structure.
A link between two filters.
int w
agreed upon image width
int h
agreed upon image height
AVFilterContext * dst
dest filter
int format
agreed upon media format
A filter pad used for either input or output.
This structure describes decoded (raw) audio or video data.
uint8_t * data[AV_NUM_DATA_POINTERS]
pointer to the picture/channel planes.
int linesize[AV_NUM_DATA_POINTERS]
For video, a positive or negative value, which is typically indicating the size in bytes of each pict...
Descriptor that unambiguously describes how the bits of a pixel are stored in the up to 4 data planes...
Common struct for all algorithms contexts.
int minknorm
@minknorm = 0 : getMax instead
double * gauss[MAX_DIFF_ORD+1]
Used for passing data between threads.
int meta_data[MAX_META_DATA]
static const AVFilterPad colorconstancy_inputs[]
static void av_always_inline get_deriv(AVFilterContext *ctx, ThreadData *td, int ord, int dir, int src, int dst, int dim, int nb_threads)
Utility function for setting up differentiation data/metadata.
#define GAUSS(s, sr, sc, sls, sh, sw, g)
static int diagonal_transformation(AVFilterContext *ctx, void *arg, int jobnr, int nb_jobs)
Performs simple correction via diagonal transformation model.
static void normalize_light(double *light)
Normalizes estimated illumination since only illumination vector direction is required for color cons...
static int setup_derivative_buffers(AVFilterContext *ctx, ThreadData *td)
Allocates buffers used by grey edge for storing derivatives final and intermediate results.
static const AVOption greyedge_options[]
static int filter_grey_edge(AVFilterContext *ctx, AVFrame *in)
Main control function for grey edge algorithm.
static int slice_normalize(AVFilterContext *ctx, void *arg, int jobnr, int nb_jobs)
Slice Frobius normalization of gaussian derivatives.
static int filter_frame(AVFilterLink *inlink, AVFrame *in)
static int get_derivative(AVFilterContext *ctx, ThreadData *td)
Main control function for calculating gaussian derivatives.
static int config_props(AVFilterLink *inlink)
static int set_gauss(AVFilterContext *ctx)
Sets gauss filters used for calculating gauss derivatives.
static void chromatic_adaptation(AVFilterContext *ctx, AVFrame *in, AVFrame *out)
Main control function for correcting scene illumination based on estimated illumination.
static av_cold void uninit(AVFilterContext *ctx)
static void cleanup_derivative_buffers(ThreadData *td, int nb_buff, int nb_planes)
Frees up buffers used by grey edge for storing derivatives final and intermediate results.
static int illumination_estimation(AVFilterContext *ctx, AVFrame *in)
Redirects to corresponding algorithm estimation function and performs normalization after estimation.
static int slice_get_derivative(AVFilterContext *ctx, void *arg, int jobnr, int nb_jobs)
Slice calculation of gaussian derivatives.
static int filter_slice_grey_edge(AVFilterContext *ctx, void *arg, int jobnr, int nb_jobs)
Slice function for grey edge algorithm that does partial summing/maximizing of gaussian derivatives.
const AVFilterPad ff_video_default_filterpad[1]
An AVFilterPad array whose only entry has name "default" and is of type AVMEDIA_TYPE_VIDEO.
AVFrame * ff_get_video_buffer(AVFilterLink *link, int w, int h)
Request a picture buffer with a specific set of permissions.
static int slice_start(SliceContext *sc, VVCContext *s, VVCFrameContext *fc, const CodedBitstreamUnit *unit, const int is_first_slice)